cimport h3lib
from h3lib cimport bool, H3int
from .util cimport (
    check_cell,
    check_edge,
    check_res,
    create_ptr,
    create_mv,
    deg2coord,
    coord2deg,
)
from libc cimport stdlib

from .util import H3ValueError


cpdef H3int geo_to_h3(double lat, double lng, int res) except 1:
    cdef:
        h3lib.GeoCoord c

    check_res(res)

    c = deg2coord(lat, lng)

    return h3lib.geoToH3(&c, res)


cpdef (double, double) h3_to_geo(H3int h) except *:
    """Map an H3 cell into its centroid geo-coordinate (lat/lng)"""
    cdef:
        h3lib.GeoCoord c

    check_cell(h)

    h3lib.h3ToGeo(h, &c)

    return coord2deg(c)


cdef h3lib.Geofence make_geofence(geos, bool lnglat_order=False) except *:
    """

    The returned `Geofence` must be freed with a call to `free_geofence`.

    Parameters
    ----------
    geos : list or tuple
        GeoFence: A sequence of >= 3 (lat, lng) pairs where the last
        element may or may not be same as the first (to form a closed loop).
        The order of the pairs may be either clockwise or counterclockwise.
    lnglat_order : bool
        If True, assume coordinate pairs like (lng, lat)
        If False, assume coordinate pairs like (lat, lng)
    """
    cdef:
        h3lib.Geofence gf

    gf.numVerts = len(geos)

    gf.verts = <h3lib.GeoCoord*> stdlib.calloc(gf.numVerts, sizeof(h3lib.GeoCoord))

    if lnglat_order:
        latlng = (g[::-1] for g in geos)
    else:
        latlng = geos

    for i, (lat, lng) in enumerate(latlng):
        gf.verts[i] = deg2coord(lat, lng)

    return gf


cdef free_geofence(h3lib.Geofence* gf):
    stdlib.free(gf.verts)
    gf.verts = NULL


cdef class GeoPolygon:
    cdef:
        h3lib.GeoPolygon gp

    def __cinit__(self, outer, holes=None, bool lnglat_order=False):
        """

        Parameters
        ----------
        outer : list or tuple
            GeoFence
            A GeoFence is a sequence of >= 3 (lat, lng) pairs where the last
            element may or may not be same as the first (to form a closed loop).
            The order of the pairs may be either clockwise or counterclockwise.
        holes : list or tuple
            A sequence of GeoFences
        lnglat_order : bool
        If True, assume coordinate pairs like (lng, lat)
        If False, assume coordinate pairs like (lat, lng)

        """
        if holes is None:
            holes = []

        self.gp.geofence = make_geofence(outer, lnglat_order)
        self.gp.numHoles = len(holes)
        self.gp.holes = NULL

        if len(holes) > 0:
            self.gp.holes =  <h3lib.Geofence*> stdlib.calloc(len(holes), sizeof(h3lib.Geofence))
            for i, hole in enumerate(holes):
                self.gp.holes[i] = make_geofence(hole, lnglat_order)


    def __dealloc__(self):
        free_geofence(&self.gp.geofence)

        for i in range(self.gp.numHoles):
            free_geofence(&self.gp.holes[i])

        stdlib.free(self.gp.holes)


def polyfill_polygon(outer, int res, holes=None, bool lnglat_order=False):
    """ Set of hexagons whose center is contained in a polygon.

    The polygon is defined as in the GeoJson standard, with an exterior
    LinearRing `outer` and a list of LinearRings `holes`, which define any
    holes in the polygon.

    Each LinearRing may be in clockwise or counter-clockwise order
    (right-hand rule or not), and may or may not be a closed loop (where the last
    element is equal to the first).
    The GeoJSON spec requires the right-hand rule, and a closed loop, but
    this function will work with any input format.

    Parameters
    ----------
    outer : list or tuple
        A LinearRing, a sequence of (lat/lng) or (lng/lat) pairs
    res : int
        The resolution of the output hexagons
    holes : list or tuple
        A collection of LinearRings, describing any holes in the polygon
    lnglat_order : bool
        If True, assume coordinate pairs like (lng, lat)
        If False, assume coordinate pairs like (lat, lng)
    """

    check_res(res)
    gp = GeoPolygon(outer, holes=holes, lnglat_order=lnglat_order)

    n = h3lib.maxPolyfillSize(&gp.gp, res)
    ptr = create_ptr(n)

    h3lib.polyfill(&gp.gp, res, ptr)
    mv = create_mv(ptr, n)

    return mv


def polyfill_geojson(geojson, int res):
    """ Set of hexagons whose center is contained in a GeoJson Polygon object.

    The polygon is defined exactly as in the GeoJson standard, so
    `geojson` should be a dictionary like:
    {
        'type': 'Polygon',
        'coordinates': [...]
    }

    'coordinates' should be a list of LinearRings, where the first ring describes
    the exterior boundary of the Polygon, and any subsequent LinearRings
    describe holes in the polygon.

    Note that we don't provide an option for the order of the coordinates,
    as the GeoJson standard requires them to be in lng/lat order.

    Parameters
    ----------
    geojson : dict
    res : int
        The resolution of the output hexagons
    """

    # todo: this one could handle multipolygons...

    if geojson['type'] != 'Polygon':
        raise ValueError('Only Polygon GeoJSON supported')

    coords = geojson['coordinates']

    out = polyfill_polygon(coords[0], res, holes=coords[1:], lnglat_order=True)

    return out


def polyfill(dict geojson, int res, bool geo_json_conformant=False):
    """ Light wrapper around `polyfill_geojson` to provide backward compatibility.
    """

    try:
        gj_type = geojson['type']
    except KeyError:
        raise KeyError("`geojson` dict must have key 'type'.") from None

    if gj_type != 'Polygon':
        raise ValueError('Only Polygon GeoJSON supported')

    if geo_json_conformant:
        out = polyfill_geojson(geojson, res)
    else:
        coords = geojson['coordinates']
        out = polyfill_polygon(coords[0], res, holes=coords[1:], lnglat_order=False)

    return out


def cell_boundary(H3int h, bool geo_json=False):
    """Compose an array of geo-coordinates that outlines a hexagonal cell"""
    cdef:
        h3lib.GeoBoundary gb

    check_cell(h)

    h3lib.h3ToGeoBoundary(h, &gb)

    verts = tuple(
        coord2deg(gb.verts[i])
        for i in range(gb.num_verts)
    )

    if geo_json:
        #lat/lng -> lng/lat and last point same as first
        verts += (verts[0],)
        verts = tuple(v[::-1] for v in verts)

    return verts


def edge_boundary(H3int edge, bool geo_json=False):
    """ Returns the GeoBoundary containing the coordinates of the edge
    """
    cdef:
        h3lib.GeoBoundary gb

    check_edge(edge)

    h3lib.getH3UnidirectionalEdgeBoundary(edge, &gb)

    # todo: move this verts transform into the GeoBoundary object
    verts = tuple(
        coord2deg(gb.verts[i])
        for i in range(gb.num_verts)
    )

    if geo_json:
        #lat/lng -> lng/lat and last point same as first
        verts += (verts[0],)
        verts = tuple(v[::-1] for v in verts)

    return verts


cpdef double point_dist(
    double lat1, double lng1,
    double lat2, double lng2, unit='km') except -1:

    a = deg2coord(lat1, lng1)
    b = deg2coord(lat2, lng2)

    if unit == 'rads':
        d = h3lib.pointDistRads(&a, &b)
    elif unit == 'km':
        d = h3lib.pointDistKm(&a, &b)
    elif unit == 'm':
        d = h3lib.pointDistM(&a, &b)
    else:
        raise H3ValueError('Unknown unit: {}'.format(unit))

    return d



import h3
import itertools
import pytest


def reverse(loop):
    return list(reversed(loop))


def drop_last(loop):
    return loop[:-1]


def toggle_map(func, poly):
    """ Return all permuations of `func` being applied or not
    to each element of `poly`

    returns iterable of length 2**len(poly)
    """
    mapped = (list(func(loop)) for loop in poly)

    return itertools.product(*zip(poly, mapped))


def chain_toggle_map(func, seq):
    seq = (toggle_map(func, p) for p in seq)
    seq = itertools.chain(*seq)

    return seq


def input_permutations(poly, res=5):
    g = [poly]
    g = chain_toggle_map(drop_last, g)
    g = chain_toggle_map(reverse, g)

    for p in g:
        hexes = h3.polyfill_polygon(p[0], res=res, holes=p[1:])
        yield hexes


def swap_element_order(seq):
    return [e[::-1] for e in seq]


def get_us_box_coords(order='latlng'):

    # big center chunk of the US in lat/lng order
    outer = [
        [42.68, -110.61],
        [32.17, -109.02],
        [31.57,  -94.26],
        [42.94,  -89.38],
        [42.68, -110.61]
    ]

    hole1 = [
        [39.77, -105.07],
        [34.81, -104.72],
        [34.77,  -98.39],
        [40.14,  -96.72],
        [39.77, -105.07]
    ]

    hole2 = [
        [41.37, -98.61],
        [40.04, -91.80],
        [42.32, -91.80],
        [41.37, -98.61]
    ]

    if order == 'lnglat':
        outer, hole1, hole2 = map(swap_element_order, [outer, hole1, hole2])

    return outer, hole1, hole2


def test_polyfill_polygon():

    # lat/lngs for State of Maine
    maine = [
        (45.137451890638886, -67.13734351262877),
        (44.8097, -66.96466),
        (44.3252, -68.03252),
        (43.98, -69.06),
        (43.68405, -70.11617),
        (43.090083319667144, -70.64573401557249),
        (43.08003225358635, -70.75102474636725),
        (43.21973948828747, -70.79761105007827),
        (43.36789581966826, -70.98176001655037),
        (43.46633942318431, -70.94416541205806),
        (45.3052400000002, -71.08482),
        (45.46022288673396, -70.6600225491012),
        (45.914794623389355, -70.30495378282376),
        (46.69317088478567, -70.00014034695016),
        (47.44777598732787, -69.23708614772835),
        (47.184794623394396, -68.90478084987546),
        (47.35462921812177, -68.23430497910454),
        (47.066248887716995, -67.79035274928509),
        (45.702585354182816, -67.79141211614706),
        (45.137451890638886, -67.13734351262877)
    ]

    # a very rough hexagonal approximation to the State of Maine
    expected = {
        '832b13fffffffff',
        '832b18fffffffff',
        '832b1afffffffff',
        '832b1efffffffff',
        '832ba9fffffffff',
        '832badfffffffff'
    }

    out = h3.polyfill_polygon(maine, 3)

    assert out == expected


def test_polyfill_polygon_order():
    lnglat, _, _ = get_us_box_coords(order='lnglat')

    out = h3.polyfill_polygon(lnglat, 5, lnglat_order=True)

    assert len(out) == 7063


# # todo: we can generate segfaults with malformed input data to polyfill
# # need to test for this and avoid segfault
# # def test_polyfill_segfault():
# #     pass


def test_polyfill_polygon_holes():

    outer, hole1, hole2 = get_us_box_coords()

    assert 7063 == len(
        h3.polyfill_polygon(outer, 5)
    )

    for res in 1, 2, 3, 4, 5:
        hexes_all = h3.polyfill_polygon(outer, res)
        hexes_holes = h3.polyfill_polygon(outer, res, [hole1, hole2])

        hexes_1 = h3.polyfill_polygon(hole1, res)
        hexes_2 = h3.polyfill_polygon(hole2, res)

        assert len(hexes_all) == len(hexes_holes) + len(hexes_1) + len(hexes_2)
        assert hexes_all == set.union(hexes_holes, hexes_1, hexes_2)


def test_polyfill_geojson():
    outer, hole1, hole2 = get_us_box_coords(order='lnglat')

    d = {
        'type': 'Polygon',
        'coordinates': [outer],
    }

    out = h3.polyfill_geojson(d, 5)

    assert len(out) == 7063


def test_polyfill():
    outer, hole1, hole2 = get_us_box_coords(order='lnglat')

    d = {
        'type': 'Polygon',
        'coordinates': [outer],
    }

    out = h3.polyfill(d, 5, geo_json_conformant=True)

    assert len(out) == 7063


def test_input_format():
    """ Test that `polyfill_polygon` can take in polygon inputs
    where the LinearRings may or may not follow the right hand rule,
    and they may or may not be closed loops (where the last element
    is equal to the first).

    Test all permutations of these rules on polygons with
    0, 1, and 2 holes. Ensure that for any polygon, each LinearRing
    may follow a different subset of rules.
    """

    poly = get_us_box_coords(order='latlng')

    assert len(poly) == 3

    # two holes
    for hexes in input_permutations(poly[:3]):
        assert len(hexes) == 5437

    # one hole
    for hexes in input_permutations(poly[:2]):
        assert len(hexes) == 5726

    # zero holes
    for hexes in input_permutations(poly[:1]):
        assert len(hexes) == 7063


def test_resolution():
    d = {
        'type': 'Polygon',
        'coordinates': [[]],
    }

    with pytest.raises(h3.H3ResolutionError):
        h3.polyfill(d, -1)

    with pytest.raises(h3.H3ResolutionError):
        h3.polyfill(d, 16)


def test_invalid_polygon():
    """
    We were previously seeing segfaults on data like
    this because we weren't raising errors inside
    some `cdef` functions.
    """

    # one
    d = {
        'type': 'Polygon',
        'coordinates': [1, 2, 3],
    }
    with pytest.raises(TypeError):
        h3.polyfill(d, 4)

    # two
    d = {
        'type': 'Polygon',
        'coordinates': [[1, 2, 3]],
    }
    with pytest.raises(TypeError):
        h3.polyfill(d, 4)

    # three
    d = {
        'type': 'Polygon',
        'coordinates': [(1, 2), (2, 2), (2, 1), (1, 2)],
    }
    with pytest.raises(TypeError):
        h3.polyfill(d, 4)



from libc cimport stdlib
from cython.view cimport array
from .h3lib cimport H3int, H3str, h3IsValid, h3UnidirectionalEdgeIsValid

cimport h3lib


cdef h3lib.GeoCoord deg2coord(double lat, double lng) nogil:
    cdef:
        h3lib.GeoCoord c

    c.lat = h3lib.degsToRads(lat)
    c.lng = h3lib.degsToRads(lng)

    return c


cdef (double, double) coord2deg(h3lib.GeoCoord c) nogil:
    return (
        h3lib.radsToDegs(c.lat),
        h3lib.radsToDegs(c.lng)
    )


cpdef basestring c_version():
    v = (
        h3lib.H3_VERSION_MAJOR,
        h3lib.H3_VERSION_MINOR,
        h3lib.H3_VERSION_PATCH,
    )

    return '{}.{}.{}'.format(*v)


cpdef H3int hex2int(H3str h) except? 0:
    return int(h, 16)


cpdef H3str int2hex(H3int x):
    """ Convert H3 integer to hex string representation

    Need to be careful in Python 2 because `hex(x)` may return a string
    with a trailing `L` character (denoting a "large" integer).
    The formatting approach below avoids this.

    Also need to be careful about unicode/str differences.
    """
    return '{:x}'.format(x)


class H3ValueError(ValueError):
    pass

class H3CellError(H3ValueError):
    pass

class H3EdgeError(H3ValueError):
    pass

class H3ResolutionError(H3ValueError):
    pass

class H3DistanceError(H3ValueError):
    pass

cdef check_cell(H3int h):
    """ Check if valid H3 "cell" (hexagon or pentagon).

    Does not check if a valid H3 edge, for example.

    Since this function is used by multiple interfaces (int or str),
    we want the error message to be informative to the user
    in either case.

    We use the builtin `hex` function instead of `int2hex` to
    prepend `0x` to indicate that this **integer** representation
    is incorrect, but in a format that is easily compared to
    `str` inputs.
    """
    if h3IsValid(h) == 0:
        raise H3CellError('Integer is not a valid H3 cell: {}'.format(hex(h)))

cdef check_edge(H3int e):
    if h3UnidirectionalEdgeIsValid(e) == 0:
        raise H3EdgeError('Integer is not a valid H3 edge: {}'.format(hex(e)))

cdef check_res(int res):
    if (res < 0) or (res > 15):
        raise H3ResolutionError(res)

cdef check_distance(int k):
    if k < 0:
        raise H3DistanceError(
            'Grid distances must be nonnegative. Received: {}'.format(k)
        )


## todo: can i turn these two into a context manager?
cdef H3int* create_ptr(size_t n) except? NULL:
    cdef H3int* ptr = <H3int*> stdlib.calloc(n, sizeof(H3int))
    if (n > 0) and (not ptr):
        raise MemoryError()

    return ptr


cdef H3int[:] create_mv(H3int* ptr, size_t n):
    cdef:
        array x

    n = move_nonzeros(ptr, n)
    if n <= 0:
        stdlib.free(ptr)
        return empty_memory_view()

    ptr = <H3int*> stdlib.realloc(ptr, n*sizeof(H3int))

    if ptr is NULL:
        raise MemoryError()

    x = <H3int[:n]> ptr
    x.callback_free_data = stdlib.free

    return x


cpdef H3int[:] from_iter(hexes):
    """ hexes needs to be an iterable that knows its size...
    or should we have it match the np.fromiter function, which infers if not available?
    """
    cdef:
        array x
        size_t n
    n = len(hexes)

    if n == 0:
        return empty_memory_view()

    x = <H3int[:n]> create_ptr(n)
    x.callback_free_data = stdlib.free

    for i,h in enumerate(hexes):
        x[i] = h

    return x


cdef size_t move_nonzeros(H3int* a, size_t n):
    """ Move nonzero elements to front of array `a` of length `n`.
    Return the number of nonzero elements.

    Loop invariant: Everything *before* `i` or *after* `j` is "done".
    Move `i` and `j` inwards until they equal, and exit.
    You can move `i` forward until there's a zero in front of it.
    You can move `j` backward until there's a nonzero to the left of it.
    Anything to the right of `j` is "junk" that can be reallocated.

    | a | b | 0 | c | d | ... |
            ^           ^
            i           j


    | a | b | d | c | d | ... |
            ^       ^
            i       j
    """
    cdef:
        size_t i = 0
        size_t j = n

    while i < j:
        if a[j-1] == 0:
            j -= 1
            continue

        if a[i] != 0:
            i += 1
            continue

        # if we're here, we know:
        # a[i] == 0
        # a[j-1] != 0
        # i < j
        # so we can swap! (actually, move a[j-1] -> a[i])
        a[i] = a[j-1]
        j -= 1

    return i


cdef H3int[:] empty_memory_view():
    # there's gotta be a better way to do this...
    # create an empty cython.view.array?
    cdef:
        H3int a[1]

    return (<H3int[:]>a)[:0]



cimport h3lib
from .h3lib cimport bool, H3int

from .util cimport (
    check_cell,
    check_edge,
    check_res,
    create_ptr,
    create_mv,
)

from .util import H3ValueError

cpdef bool are_neighbors(H3int h1, H3int h2):
    check_cell(h1)
    check_cell(h2)

    return h3lib.h3IndexesAreNeighbors(h1, h2) == 1


cpdef H3int edge(H3int origin, H3int destination) except 1:
    check_cell(origin)
    check_cell(destination)

    if h3lib.h3IndexesAreNeighbors(origin, destination) != 1:
        s = 'Cells are not neighbors: {} and {}'
        s = s.format(hex(origin), hex(destination))
        raise H3ValueError(s)

    return h3lib.getH3UnidirectionalEdge(origin, destination)


cpdef bool is_edge(H3int e):
    return h3lib.h3UnidirectionalEdgeIsValid(e) == 1

cpdef H3int edge_origin(H3int e) except 1:
    # without the check, with an invalid input, the function will just return 0
    check_edge(e)

    return h3lib.getOriginH3IndexFromUnidirectionalEdge(e)

cpdef H3int edge_destination(H3int e) except 1:
    check_edge(e)

    return h3lib.getDestinationH3IndexFromUnidirectionalEdge(e)

cpdef (H3int, H3int) edge_cells(H3int e) except *:
    check_edge(e)

    return edge_origin(e), edge_destination(e)

cpdef H3int[:] edges_from_cell(H3int origin):
    """ Returns the 6 (or 5 for pentagons) directed edges
    for the given origin cell
    """
    check_cell(origin)

    ptr = create_ptr(6)
    h3lib.getH3UnidirectionalEdgesFromHexagon(origin, ptr)
    mv = create_mv(ptr, 6)

    return mv


cpdef double mean_edge_length(int resolution, unit='km') except -1:
    check_res(resolution)

    length = h3lib.edgeLengthKm(resolution)

    # todo: multiple units
    convert = {
        'km': 1.0,
        'm': 1000.0
    }

    try:
        length *= convert[unit]
    except:
        raise H3ValueError('Unknown unit: {}'.format(unit))

    return length


cpdef double edge_length(H3int e, unit='km') except -1:
    check_edge(e)

    # todo: maybe kick this logic up to the python level
    # it might be a little cleaner, because we can do the "switch statement"
    # with a dict, but would require exposing more C functions

    if unit == 'rads':
        length = h3lib.exactEdgeLengthRads(e)
    elif unit == 'km':
        length = h3lib.exactEdgeLengthKm(e)
    elif unit == 'm':
        length = h3lib.exactEdgeLengthM(e)
    else:
        raise H3ValueError('Unknown unit: {}'.format(unit))

    return length



from libc cimport stdint
from cpython cimport bool
from libc.stdint cimport int64_t

ctypedef stdint.uint64_t H3int
ctypedef basestring H3str

cdef extern from "h3api.h":
    cdef int H3_VERSION_MAJOR
    cdef int H3_VERSION_MINOR
    cdef int H3_VERSION_PATCH

    ctypedef stdint.uint64_t H3Index

    ctypedef struct GeoCoord:
        double lat  # in radians
        double lng "lon"  # in radians

    ctypedef struct GeoBoundary:
        int num_verts "numVerts"
        GeoCoord verts[10]  # MAX_CELL_BNDRY_VERTS

    ctypedef struct Geofence:
        int numVerts
        GeoCoord *verts

    ctypedef struct GeoPolygon:
        Geofence geofence
        int numHoles
        Geofence *holes

    ctypedef struct GeoMultiPolygon:
        int numPolygons
        GeoPolygon *polygons

    ctypedef struct LinkedGeoCoord:
        GeoCoord data "vertex"
        LinkedGeoCoord *next

    # renaming these for clarity
    ctypedef struct LinkedGeoLoop:
        LinkedGeoCoord *data "first"
        LinkedGeoCoord *_data_last "last"  # not needed in Cython bindings
        LinkedGeoLoop *next

    ctypedef struct LinkedGeoPolygon:
        LinkedGeoLoop *data "first"
        LinkedGeoLoop *_data_last "last"  # not needed in Cython bindings
        LinkedGeoPolygon *next

    ctypedef struct CoordIJ:
        int i
        int j

    H3Index geoToH3(const GeoCoord *g, int res) nogil

    void h3ToGeo(H3Index h3, GeoCoord *g) nogil

    void h3ToGeoBoundary(H3Index h3, GeoBoundary *gp)

    int maxKringSize(int k)

    int hexRange(H3Index origin, int k, H3Index *out)

    int hexRangeDistances(H3Index origin, int k, H3Index *out, int *distances)

    int h3Distance(H3Index origin, H3Index h3)

    int hexRanges(H3Index *h3Set, int length, int k, H3Index *out)

    void kRing(H3Index origin, int k, H3Index *out)

    void kRingDistances(H3Index origin, int k, H3Index *out, int *distances)

    int hexRing(H3Index origin, int k, H3Index *out)

    int maxPolyfillSize(const GeoPolygon *geoPolygon, int res)

    void polyfill(const GeoPolygon *geoPolygon, int res, H3Index *out)

    void h3SetToLinkedGeo(const H3Index *h3Set, const int numHexes, LinkedGeoPolygon *out)

    void destroyLinkedPolygon(LinkedGeoPolygon *polygon)

    double degsToRads(double degrees) nogil

    double radsToDegs(double radians) nogil

    stdint.int64_t numHexagons(int res)

    int h3GetResolution(H3Index h)

    int h3GetBaseCell(H3Index h)

    H3Index stringToH3(const char *str)

    void h3ToString(H3Index h, char *str, size_t sz)

    int h3IsValid(H3Index h)

    H3Index h3ToParent(H3Index h, int parentRes)

    int maxH3ToChildrenSize(H3Index h, int childRes)

    void h3ToChildren(H3Index h, int childRes, H3Index *children)

    int compact(const H3Index *h3Set, H3Index *compactedSet, const int numHexes)

    int maxUncompactSize(const H3Index *compactedSet, const int numHexes, const int res)

    int uncompact(const H3Index *compactedSet, const int numHexes, H3Index *h3Set, const int maxHexes, const int res)

    int h3IsResClassIII(H3Index h)

    int h3IsPentagon(H3Index h)

    int pentagonIndexCount()

    void getPentagonIndexes(int res, H3Index *out)

    int res0IndexCount()

    void getRes0Indexes(H3Index *out)

    H3Index h3ToCenterChild(H3Index h, int res)

    int h3IndexesAreNeighbors(H3Index origin, H3Index destination)

    H3Index getH3UnidirectionalEdge(H3Index origin, H3Index destination)

    int h3UnidirectionalEdgeIsValid(H3Index edge)

    H3Index getOriginH3IndexFromUnidirectionalEdge(H3Index edge)

    H3Index getDestinationH3IndexFromUnidirectionalEdge(H3Index edge)

    void getH3IndexesFromUnidirectionalEdge(H3Index edge, H3Index *originDestination)

    void getH3UnidirectionalEdgesFromHexagon(H3Index origin, H3Index *edges)

    void getH3UnidirectionalEdgeBoundary(H3Index edge, GeoBoundary *gb)

    int h3LineSize(H3Index start, H3Index end)
    int h3Line(H3Index start, H3Index end, H3Index *out)

    int maxFaceCount(H3Index h3)
    void h3GetFaces(H3Index h3, int *out)

    int experimentalH3ToLocalIj(H3Index origin, H3Index h3, CoordIJ *out)
    int experimentalLocalIjToH3(H3Index origin, const CoordIJ *ij, H3Index *out)

    double hexAreaKm2(int res) nogil
    double hexAreaM2(int res) nogil

    double cellAreaRads2(H3Index h) nogil
    double cellAreaKm2(H3Index h) nogil
    double cellAreaM2(H3Index h) nogil

    double edgeLengthKm(int res) nogil
    double edgeLengthM(int res) nogil

    double exactEdgeLengthRads(H3Index edge) nogil
    double exactEdgeLengthKm(H3Index edge) nogil
    double exactEdgeLengthM(H3Index edge) nogil

    double pointDistRads(const GeoCoord *a, const GeoCoord *b) nogil
    double pointDistKm(const GeoCoord *a, const GeoCoord *b) nogil
    double pointDistM(const GeoCoord *a, const GeoCoord *b) nogil



import h3


def test_h3_set_to_multi_polygon():
    h = '8928308280fffff'
    hexes = h3.k_ring(h, 1)

    mpoly = h3.h3_set_to_multi_polygon(hexes)

    out = h3.polyfill_polygon(mpoly[0][0], 9, holes=None, lnglat_order=False)

    assert out == hexes


def test_2_polys():
    h = '8928308280fffff'
    hexes = h3.hex_ring(h, 2)
    hexes = hexes | {h}
    # hexes should be a center hex, and the 2-ring around it
    # (with the 1-ring being absent)

    out = [
        h3.polyfill_polygon(poly[0], 9, holes=poly[1:], lnglat_order=False)
        for poly in h3.h3_set_to_multi_polygon(hexes, geo_json=False)
    ]

    assert set.union(*out) == hexes


def test_2_polys_json():
    h = '8928308280fffff'
    hexes = h3.hex_ring(h, 2)
    hexes = hexes | {h}
    # hexes should be a center hex, and the 2-ring around it
    # (with the 1-ring being absent)

    # not deterministic which poly is first..
    poly1, poly2 = h3.h3_set_to_multi_polygon(hexes, geo_json=True)

    assert {len(poly1), len(poly2)} == {1, 2}

    for poly in poly1, poly2:
        for loop in poly:
            assert loop[0] == loop[-1]


def test_2_polys_not_json():
    h = '8928308280fffff'
    hexes = h3.hex_ring(h, 2)
    hexes = hexes | {h}
    # hexes should be a center hex, and the 2-ring around it
    # (with the 1-ring being absent)

    # not deterministic which poly is first..
    poly1, poly2 = h3.h3_set_to_multi_polygon(hexes, geo_json=False)

    assert {len(poly1), len(poly2)} == {1, 2}

    for poly in poly1, poly2:
        for loop in poly:
            assert loop[0] != loop[-1]



import numpy as np  # only run this test suite if numpy is installed
import pytest
from h3.api import (
    basic_int,
    numpy_int,
    memview_int,
)

# todo: check when a copy is made, and when it isn't


def test_set():
    ints = {
        619056821839331327,
        619056821839593471,
        619056821839855615,
        619056821840117759,
        619056821840379903,
        619056821840642047,
        619056821840904191,
    }

    h = 614553222213795839

    assert basic_int.compact(ints) == {h}

    with pytest.raises(TypeError):
        # numpy can't convert from a set
        numpy_int.compact(ints)

    with pytest.raises(TypeError):
        # set isn't a memoryview
        memview_int.compact(ints)


def test_list():
    ints = [
        619056821839331327,
        619056821839593471,
        619056821839855615,
        619056821840117759,
        619056821840379903,
        619056821840642047,
        619056821840904191,
    ]

    h = 614553222213795839

    assert basic_int.compact(ints) == {h}

    # numpy can convert from a list OK
    # (numpy knows to convert it to uint64)
    assert numpy_int.compact(ints) == np.array([h], dtype='uint64')

    # little weird that numpy comparisons don't consider dtype
    assert numpy_int.compact(ints) == np.array([h])
    assert not numpy_int.compact(ints).dtype == np.array([h]).dtype

    with pytest.raises(TypeError):
        # list isn't a memoryview
        memview_int.compact(ints)


def test_np_array():
    ints = np.array([
        619056821839331327,
        619056821839593471,
        619056821839855615,
        619056821840117759,
        619056821840379903,
        619056821840642047,
        619056821840904191,
    ], dtype='uint64')

    h = 614553222213795839

    assert basic_int.compact(ints) == {h}

    assert numpy_int.compact(ints) == np.array([h], dtype='uint64')
    assert numpy_int.compact(ints).dtype == np.dtype('uint64')

    out = memview_int.compact(ints)
    assert len(out) == 1
    assert out[0] == h


def test_list_to_array():
    ints0 = [
        619056821839331327,
        619056821839593471,
        619056821839855615,
        619056821840117759,
        619056821840379903,
        619056821840642047,
        619056821840904191,
    ]

    ints = np.array(ints0)

    h = 614553222213795839

    assert basic_int.compact(ints) == {h}
    assert numpy_int.compact(ints) == np.array([h], dtype='uint64')

    with pytest.raises(ValueError):
        # Without the explicit dtype given above, the array
        # assumes it has *signed* integers
        # The `memview_int` interface requires a dtype match
        # with uint64.
        memview_int.compact(ints)


def test_iterator():

    def foo():
        ints = iter([
            619056821839331327,
            619056821839593471,
            619056821839855615,
            619056821840117759,
            619056821840379903,
            619056821840642047,
            619056821840904191,
        ])

        return ints

    h = 614553222213795839

    ints = foo()
    assert basic_int.compact(ints) == {h}

    ints = foo()
    with pytest.raises(TypeError):
        # numpy can't create an array from an iterator
        numpy_int.compact(ints)

    ints = foo()
    with pytest.raises(TypeError):
        # requires a bytes-like input
        memview_int.compact(ints)



cimport h3lib
from h3lib cimport H3int
from .util cimport deg2coord

from cython cimport boundscheck, wraparound
from libc.math cimport sqrt, sin, cos, asin

cdef double haversineDistance(double th1, double ph1, double th2, double ph2) nogil:
    cdef:
        double dx, dy, dz
        double R = 6371.0088

    ph1 -= ph2

    dz = sin(th1) - sin(th2)
    dx = cos(ph1) * cos(th1) - cos(th2)
    dy = sin(ph1) * cos(th1)

    return asin(sqrt(dx*dx + dy*dy + dz*dz) / 2)*2*R


@boundscheck(False)
@wraparound(False)
cpdef void haversine_vect(
    const H3int[:] a,
    const H3int[:] b,
         double[:] out
) nogil:

    cdef h3lib.GeoCoord p1, p2

    with nogil:
        # todo: add these back in when cython 3.0 comes out
        #assert len(a) == len(b)
        #assert len(a) <= len(out)

        for i in range(len(a)):
            h3lib.h3ToGeo(a[i], &p1)
            h3lib.h3ToGeo(b[i], &p2)
            out[i] = haversineDistance(
                p1.lat, p1.lng,
                p2.lat, p2.lng
            )

@boundscheck(False)
@wraparound(False)
cpdef void geo_to_h3_vect(
    const double[:] lat,
    const double[:] lng,
    int res,
    H3int[:] out
) nogil:

    cdef h3lib.GeoCoord c

    with nogil:
        for i in range(len(lat)):
            c = deg2coord(lat[i], lng[i])
            out[i] = h3lib.geoToH3(&c, res)



from .h3lib cimport H3int, H3str, GeoCoord

cdef GeoCoord deg2coord(double lat, double lng) nogil
cdef (double, double) coord2deg(GeoCoord c) nogil

cpdef H3int hex2int(H3str h) except? 0
cpdef H3str int2hex(H3int x)

cdef H3int* create_ptr(size_t n) except? NULL
cdef H3int[:] create_mv(H3int* ptr, size_t n)


cdef check_cell(H3int h)
cdef check_edge(H3int e)
cdef check_res(int res)
cdef check_distance(int k)

cdef H3int[:] empty_memory_view() # want to drop this if possible



cimport h3lib
from .h3lib cimport bool, int64_t, H3int
from libc cimport stdlib

from .util cimport (
    check_cell,
    check_res,
    check_distance,
    create_ptr,
    create_mv,
    empty_memory_view, # want to drop this import if possible
)

from .util import H3ValueError, H3ResolutionError

# todo: add notes about Cython exception handling


# bool is a python type, so we don't need the except clause
cpdef bool is_cell(H3int h):
    """Validates an H3 cell (hexagon or pentagon)

    Returns
    -------
    boolean
    """
    return h3lib.h3IsValid(h) == 1

cpdef bool is_pentagon(H3int h):
    return h3lib.h3IsPentagon(h) == 1

cpdef int get_base_cell(H3int h) except -1:
    check_cell(h)

    return h3lib.h3GetBaseCell(h)


cpdef int resolution(H3int h) except -1:
    """Returns the resolution of an H3 Index
    0--15
    """
    check_cell(h)

    return h3lib.h3GetResolution(h)


cpdef int distance(H3int h1, H3int h2) except -1:
    """ compute the hex-distance between two hexagons
    """
    check_cell(h1)
    check_cell(h2)

    d = h3lib.h3Distance(h1,h2)

    if d < 0:
        s = 'Cells are too far apart to compute distance: {} and {}'
        s = s.format(hex(h1), hex(h2))
        raise H3ValueError(s)

    return d

cpdef H3int[:] disk(H3int h, int k):
    """ Return cells at grid distance `<= k` from `h`.
    """
    check_cell(h)
    check_distance(k)

    n = h3lib.maxKringSize(k)

    ptr = create_ptr(n) # todo: return a "smart" pointer that knows its length?
    h3lib.kRing(h, k, ptr)
    mv = create_mv(ptr, n)

    return mv


cpdef H3int[:] _ring_fallback(H3int h, int k):
    """
    `ring` tries to call `h3lib.hexRing` first; if that fails, we call
    this function, which relies on `h3lib.kRingDistances`.

    Failures for `h3lib.hexRing` happen when the algortihm runs into a pentagon.
    """
    check_cell(h)
    check_distance(k)

    n = h3lib.maxKringSize(k)
    # array of h3 cells
    ptr = create_ptr(n)

    # array of cell distances from `h`
    dist_ptr = <int*> stdlib.calloc(n, sizeof(int))
    if dist_ptr is NULL:
        raise MemoryError()

    h3lib.kRingDistances(h, k, ptr, dist_ptr)

    distances = <int[:n]> dist_ptr
    distances.callback_free_data = stdlib.free

    for i,v in enumerate(distances):
        if v != k:
            ptr[i] = 0

    mv = create_mv(ptr, n)

    return mv

cpdef H3int[:] ring(H3int h, int k):
    """ Return cells at grid distance `== k` from `h`.
    Collection is "hollow" for k >= 1.
    """
    check_cell(h)
    check_distance(k)

    n = 6*k if k > 0 else 1
    ptr = create_ptr(n)

    flag = h3lib.hexRing(h, k, ptr)

    # if we drop into the failure state, we might be tempted to not create
    # this mv, but creating the mv is exactly what guarantees that we'll free
    # the memory. context manager would be better here, if we can figure out
    # how to do that
    mv = create_mv(ptr, n)

    if flag != 0:
        mv = _ring_fallback(h, k)

    return mv


cpdef H3int parent(H3int h, res=None) except 0:
    check_cell(h)

    if res is None:
        res = resolution(h) - 1
    if res > resolution(h):
        msg = 'Invalid parent resolution {} for cell {}.'
        msg = msg.format(res, hex(h))
        raise H3ResolutionError(msg)

    check_res(res)

    return h3lib.h3ToParent(h, res)

cpdef H3int[:] children(H3int h, res=None):
    check_cell(h)

    if res is None:
        res = resolution(h) + 1
    if res < resolution(h):
        msg = 'Invalid child resolution {} for cell {}.'
        msg = msg.format(res, hex(h))
        raise H3ResolutionError(msg)

    check_res(res)

    n = h3lib.maxH3ToChildrenSize(h, res)

    ptr = create_ptr(n)
    h3lib.h3ToChildren(h, res, ptr)
    mv = create_mv(ptr, n)

    return mv

cpdef H3int center_child(H3int h, res=None) except 0:
    check_cell(h)

    if res is None:
        res = resolution(h) + 1
    if res < resolution(h):
        msg = 'Invalid child resolution {} for cell {}.'
        msg = msg.format(res, hex(h))
        raise H3ResolutionError(msg)

    check_res(res)

    return h3lib.h3ToCenterChild(h, res)



cpdef H3int[:] compact(const H3int[:] hu):
    # todo: the Clib can handle 0-len arrays because it **avoids**
    # dereferencing the pointer, but Cython's syntax of
    # `&hu[0]` **requires** a dereference. For Cython, checking for array
    # length of zero and returning early seems like the easiest solution.
    # note: open to better ideas!
    if len(hu) == 0:
        return empty_memory_view()

    for h in hu: ## todo: should we have an array version? would that be faster?
        check_cell(h)

    ptr = create_ptr(len(hu))
    flag = h3lib.compact(&hu[0], ptr, len(hu))
    mv = create_mv(ptr, len(hu))

    if flag != 0:
        raise H3ValueError('Could not compact set of hexagons!')

    return mv

# todo: https://stackoverflow.com/questions/50684977/cython-exception-type-for-a-function-returning-a-typed-memoryview
# apparently, memoryviews are python objects, so we don't need to do the except clause
cpdef H3int[:] uncompact(const H3int[:] hc, int res):
    # todo: the Clib can handle 0-len arrays because it **avoids**
    # dereferencing the pointer, but Cython's syntax of
    # `&hc[0]` **requires** a dereference. For Cython, checking for array
    # length of zero and returning early seems like the easiest solution.
    # note: open to better ideas!
    if len(hc) == 0:
        return empty_memory_view()

    for h in hc:
        check_cell(h)

    N = h3lib.maxUncompactSize(&hc[0], len(hc), res)

    ptr = create_ptr(N)
    flag = h3lib.uncompact(
        &hc[0], len(hc),
           ptr, N,
        res
    )
    mv = create_mv(ptr, N)

    if flag != 0:
        raise H3ValueError('Could not uncompact set of hexagons!')

    return mv


cpdef int64_t num_hexagons(int resolution) except -1:
    check_res(resolution)

    return h3lib.numHexagons(resolution)


cpdef double mean_hex_area(int resolution, unit='km^2') except -1:
    check_res(resolution)

    area = h3lib.hexAreaKm2(resolution)

    # todo: multiple units
    convert = {
        'km^2': 1.0,
        'm^2': 1000*1000.0
    }

    try:
        area *= convert[unit]
    except:
        raise H3ValueError('Unknown unit: {}'.format(unit))

    return area


cpdef double cell_area(H3int h, unit='km^2') except -1:
    check_cell(h)

    if unit == 'rads^2':
        area = h3lib.cellAreaRads2(h)
    elif unit == 'km^2':
        area = h3lib.cellAreaKm2(h)
    elif unit == 'm^2':
        area = h3lib.cellAreaM2(h)
    else:
        raise H3ValueError('Unknown unit: {}'.format(unit))

    return area


cpdef H3int[:] line(H3int start, H3int end):
    check_cell(start)
    check_cell(end)

    n = h3lib.h3LineSize(start, end)

    if n < 0:
        s = "Couldn't find line between cells {} and {}"
        s = s.format(hex(start), hex(end))
        raise H3ValueError(s)

    ptr = create_ptr(n)
    flag = h3lib.h3Line(start, end, ptr)
    mv = create_mv(ptr, n)

    if flag != 0:
        s = "Couldn't find line between cells {} and {}"
        s = s.format(hex(start), hex(end))
        raise H3ValueError(s)

    return mv

cpdef bool is_res_class_iii(H3int h):
    return h3lib.h3IsResClassIII(h) == 1


cpdef H3int[:] get_pentagon_indexes(int res):
    check_res(res)

    n = h3lib.pentagonIndexCount()

    ptr = create_ptr(n)
    h3lib.getPentagonIndexes(res, ptr)
    mv = create_mv(ptr, n)

    return mv


cpdef H3int[:] get_res0_indexes():
    n = h3lib.res0IndexCount()

    ptr = create_ptr(n)
    h3lib.getRes0Indexes(ptr)
    mv = create_mv(ptr, n)

    return mv


cpdef get_faces(H3int h):
    check_cell(h)

    n = h3lib.maxFaceCount(h)

    cdef int* ptr = <int*> stdlib.calloc(n, sizeof(int))
    if (n > 0) and (not ptr):
        raise MemoryError()

    h3lib.h3GetFaces(h, ptr)

    faces = <int[:n]> ptr
    faces = {f for f in faces if f >= 0}
    stdlib.free(ptr)

    return faces


cpdef (int, int) experimental_h3_to_local_ij(H3int origin, H3int h) except *:
    cdef:
        int flag
        h3lib.CoordIJ c

    check_cell(origin)
    check_cell(h)

    flag = h3lib.experimentalH3ToLocalIj(origin, h, &c)

    if flag != 0:
        s = "Couldn't find local (i,j) between cells {} and {}."
        s = s.format(hex(origin), hex(h))
        raise H3ValueError(s)

    return c.i, c.j


cpdef H3int experimental_local_ij_to_h3(H3int origin, int i, int j) except 0:
    cdef:
        int flag
        h3lib.CoordIJ c
        H3int out

    check_cell(origin)

    c.i, c.j = i, j

    flag = h3lib.experimentalLocalIjToH3(origin, &c, &out)

    if flag != 0:
        s = "Couldn't find cell at local ({},{}) from cell {}."
        s = s.format(i, j, hex(origin))
        raise H3ValueError(s)

    return out



import h3.api.memview_int as h3


def test1():
    assert h3.geo_to_h3(37.7752702151959, -122.418307270836, 9) == 617700169958293503


def test_line():
    h1 = '8928308280fffff'
    h2 = '8928308287bffff'
    h1, h2 = h3.string_to_h3(h1), h3.string_to_h3(h2)

    out = h3.h3_line(h1, h2)

    # todo: are we outputting `memoryviewslice`? should we just output a memoryview?
    assert 'memoryview' in str(type(out))

    expected = [
        617700169958293503,
        617700169964847103,
        617700169965371391,
    ]

    assert list(out) == expected


def test_get_faces():
    h = 577832942814887935
    expected = {2, 3, 7, 8, 12}
    out = h3.h3_get_faces(h)
    assert out == expected

    h = 579873636396040191
    expected = {13}
    out = h3.h3_get_faces(h)
    assert out == expected

    h = 579768083279773695
    expected = {16, 15}
    out = h3.h3_get_faces(h)
    assert out == expected
